# julia Gtk interface
module Gtk

if isdefined(Base, :Experimental) && isdefined(Base.Experimental, Symbol("@optlevel"))
    @eval Base.Experimental.@optlevel 1
end

# Import binary definitions
using GTK3_jll, Glib_jll, Xorg_xkeyboard_config_jll, gdk_pixbuf_jll, adwaita_icon_theme_jll, hicolor_icon_theme_jll
using Librsvg_jll
using JLLWrappers
using Pkg.Artifacts, Scratch
const libgdk = libgdk3
const libgtk = libgtk3
const libgdk_pixbuf = libgdkpixbuf


const suffix = :Leaf
include("GLib/GLib.jl")
using .GLib
using .GLib.MutableTypes
import .GLib: set_gtk_property!, get_gtk_property, getproperty, FieldRef
import .GLib:
    signal_connect, signal_handler_disconnect,
    signal_handler_block, signal_handler_unblock, signal_handler_is_connected,
    signal_emit, unsafe_convert,
    AbstractStringLike, bytestring

import Base: convert, show, run, size, resize!, length, getindex, setindex!,
             insert!, push!, append!, pushfirst!, pop!, splice!, delete!, deleteat!,
             parent, isempty, empty!, first, last, in, popfirst!,
             eltype, copy, isvalid, string, sigatomic_begin, sigatomic_end, (:), iterate

export showall, select!, start

using Reexport
@reexport using Graphics
import .Graphics: width, height, getgc

using Cairo
import Cairo: destroy
using Serialization

const Index{I<:Integer} = Union{I, AbstractVector{I}}

export GAccessor
include("basic_exports.jl")
include("long_exports.jl")
include("long_leaf_exports.jl")

global const libgtk_version = VersionNumber(
      ccall((:gtk_get_major_version, libgtk), Cint, ()),
      ccall((:gtk_get_minor_version, libgtk), Cint, ()),
      ccall((:gtk_get_micro_version, libgtk), Cint, ()))

include("gdk.jl")
include("interfaces.jl")
include("boxes.jl")
include("gtktypes.jl")
include("base.jl")
include("events.jl")
include("container.jl")
include("layout.jl")
include("displays.jl")
include("lists.jl")
include("buttons.jl")
include("input.jl")
include("text.jl")
include("menus.jl")
include("selectors.jl")
include("misc.jl")
include("cairo.jl")
include("builder.jl")
include("toolbar.jl")
include("theme.jl")
include("gio.jl")
include("application.jl")

function __init__()
    in(:Gtk4, names(Main, imported=true)) && error("Gtk is incompatible with Gtk4.")

    # Set XDG_DATA_DIRS so that Gtk can find its icons and schemas
    ENV["XDG_DATA_DIRS"] = join(filter(x -> x !== nothing, [
        dirname(adwaita_icons_dir),
        dirname(hicolor_icons_dir),
        joinpath(dirname(GTK3_jll.libgdk3_path::String), "..", "share"),
        get(ENV, "XDG_DATA_DIRS", nothing)::Union{String,Nothing},
    ]), Sys.iswindows() ? ";" : ":")

    # Next, ensure that gdk-pixbuf has its loaders.cache file; we generate a
    # MutableArtifacts.toml file that maps in a loaders.cache we dynamically
    # generate by running `gdk-pixbuf-query-loaders:`
    cache_dir = @get_scratch!("gdk-pixbuf-cache")
    treehash_cache_path = joinpath(cache_dir, "gdk_pixbuf_treehash.cache")
    loaders_cache_path = joinpath(cache_dir, "loaders.cache")
    gdk_pixbuf_treehash = basename(gdk_pixbuf_jll.artifact_dir)
    if !isfile(treehash_cache_path) || read(treehash_cache_path) != gdk_pixbuf_treehash
        function query_pixbuf_loaders(dir::String;
                                      extra_env::Vector{Pair{String,String}} = Pair{String,String}[])
            gpql = gdk_pixbuf_query_loaders()
            return readchomp(addenv(gpql, "GDK_PIXBUF_MODULEDIR" => dir, extra_env...))
        end

        open(loaders_cache_path, write=true) do io
            # Cache builtin gdx-pixbuf modules
            write(io, query_pixbuf_loaders(gdk_pixbuf_loaders_dir))
            println(io)

            # If Librsvg_jll is available, cache that one too
            if Librsvg_jll.is_available()
                librsvg_module_dir = dirname(Librsvg_jll.libpixbufloader_svg)
                write(io, query_pixbuf_loaders(librsvg_module_dir; extra_env = [
                    JLLWrappers.LIBPATH_env=>Librsvg_jll.LIBPATH[],
                ]))
                println(io)
            end
        end

        # TODO: I think this is unnecessary
        #Sys.iswindows() && chmod(artifact_path(loaders_dir_hash), 0o755; recursive=true)

        rm(treehash_cache_path; force=true)
        open(treehash_cache_path, write=true) do io
            write(io, gdk_pixbuf_treehash)
        end
    end

    # Point gdk to our cached loaders
    ENV["GDK_PIXBUF_MODULE_FILE"] = loaders_cache_path

    # TODO: I think this is unnecessary
    #ENV["GDK_PIXBUF_MODULEDIR"] = gdk_pixbuf_loaders_dir

    if Sys.islinux() || Sys.isfreebsd()
        # Needed by xkbcommon:
        # https://xkbcommon.org/doc/current/group__include-path.html.  Related
        # to issue https://github.com/JuliaGraphics/Gtk.jl/issues/469
        ENV["XKB_CONFIG_ROOT"] = joinpath(Xorg_xkeyboard_config_jll.artifact_dir::String,
                                          "share", "X11", "xkb")
    end

    GError() do error_check
        ccall((:gtk_init_with_args, libgtk), Bool,
            (Ptr{Nothing}, Ptr{Nothing}, Ptr{UInt8}, Ptr{Nothing}, Ptr{UInt8}, Ptr{Ptr{GError}}),
            C_NULL, C_NULL, "Julia Gtk Bindings", C_NULL, C_NULL, error_check)
    end

    # if g_main_depth > 0, a glib main-loop is already running.
    # unfortunately this call does not reliably reflect the state after the
    # loop has been stopped or restarted, so only use it once at the start
    gtk_main_running[] = ccall((:g_main_depth, GLib.libglib), Cint, ()) > 0

    # Given GLib provides `g_idle_add` to specify what happens during idle, this allows
    # that call to also start the eventloop
    GLib.gtk_eventloop_f[] = enable_eventloop

    auto_idle[] = get(ENV, "GTK_AUTO_IDLE", "true") == "true"

    # by default, defer starting the event loop until either `show`, `showall`, or `g_idle_add` is called
    enable_eventloop(!auto_idle[])
end

const auto_idle = Ref{Bool}(true) # control default via ENV["GTK_AUTO_IDLE"]
const gtk_main_running = Ref{Bool}(false)
const quit_task = Ref{Task}()
const enable_eventloop_lock = Base.ReentrantLock()
"""
    Gtk.enable_eventloop(b::Bool = true)

Set whether Gtk's event loop is running.
"""
function enable_eventloop(b::Bool = true; wait_stopped::Bool = false)
    lock(enable_eventloop_lock) do # handle widgets that are being shown/destroyed from different threads
        isassigned(quit_task) && wait(quit_task[]) # prevents starting while the async is still stopping
        if b
            if !is_eventloop_running()
                global gtk_main_task = schedule(Task(gtk_main))
                gtk_main_running[] = true
            end
        else
            if is_eventloop_running()
                # @async and short sleep is needer on MacOS at least, otherwise
                # the window doesn't always finish closing before the eventloop stops.
                quit_task[] = @async begin
                    sleep(0.2)
                    gtk_quit()
                    gtk_main_running[] = false
                end
                wait_stopped && wait(quit_task[])
            end
        end
    end
end

"""
    Gtk.pause_eventloop(f; force = false)

Pauses the eventloop around a function. Restores the state of the eventloop after
pausing. Respects whether Gtk.jl is configured to allow auto-stopping of the
eventloop, unless `force = true`.
"""
function pause_eventloop(f; force = false)
    was_running = is_eventloop_running()
    (force || auto_idle[]) && enable_eventloop(false, wait_stopped = true)
    try
        f()
    finally
        (force || auto_idle[]) && enable_eventloop(was_running)
    end
end

"""
    Gtk.is_eventloop_running()::Bool

Check whether Gtk's event loop is running.
"""
is_eventloop_running() = gtk_main_running[]

const ser_version = Serialization.ser_version
let cachedir = joinpath(splitdir(@__FILE__)[1], "..", "gen")
    fastgtkcache = joinpath(cachedir, "gtk$(libgtk_version.major)_julia_ser$(ser_version)")
    if isfile(fastgtkcache) && true
        open(fastgtkcache) do cache
            while !eof(cache)
                Core.eval(Gtk, deserialize(cache))
            end
        end
    else
        gboxcache = joinpath(cachedir, "gbox$(libgtk_version.major)")
        map(eval, include(gboxcache).args)
        constcache = joinpath(cachedir, "gconsts$(libgtk_version.major)")
        map(eval, include(constcache).args)
    end
end
const _ = GAccessor
using .GConstants

include("windows.jl")
include("gl_area.jl")

if Base.VERSION >= v"1.4.2"
    include("precompile.jl")
    _precompile_()
end

# Alternative Interface (`using Gtk.ShortNames`)
module ShortNames
    using ..Gtk
    import ..GLib:
        signal_connect, signal_handler_disconnect,
        signal_handler_block, signal_handler_unblock, signal_handler_is_connected,
        signal_emit
    import ..GLib.@g_type_delegate
    import ..Gtk: suffix
    export Gtk
    include("basic_exports.jl")
    include("short_exports.jl")
    include("short_leaf_exports.jl")
end
using .ShortNames
end
